/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.CredentialManager;

import java.io.*;
import java.security.*;
import java.util.*;
import java.util.regex.Pattern;

import java.security.cert.*;
import sun.misc.BASE64Decoder;

import org.bouncycastle.jce.provider.*;
import org.bouncycastle.x509.*;


import org.opensaml.*;
import org.apache.xml.security.signature.XMLSignature;


/** Manages with digital certificates validation and verification,
 * and initializaiton of trusted CAs and SOAs. 
 * <p> For more information consult the source.
 */
public class CertificateManager {
          
       private static ArrayList SOAPublicKeys = null;
       private static ArrayList CAPublicKeys = null;
       private static BouncyCastleProvider BC_provider = null;
       private static Key SSLPrivateKey = null;
       
       
      /** The hashtable stores all credentials to be protected by unauthorized
       * access. They are called own credentials (in possession) by an opponent.
       * key - logical representation; value - path to the certificate file.*/
      private static Hashtable myCredentials = new Hashtable(); 
      public static Hashtable CredentialsValidity = new Hashtable(); 
         
    /** Creates a new instance of CertificateManager */
    public CertificateManager() {
        BC_provider = new BouncyCastleProvider();
        Security.addProvider( BC_provider );   
    }
    
    /** Loads credentials in possession by the agent.*/
    public static void loadMyCredentials() {
        try {
            String certAsString;
            
            File[] credentialsDir = (new File(CredentialsConfig.CertificateDB)).listFiles();
            for ( int y = 0; y <  credentialsDir.length; y++ ) 
            {
                if ( credentialsDir[y].getPath().endsWith(".cer") )
                {
                    try
                    {
                        FileInputStream is = new FileInputStream(credentialsDir[y]); //for files that contain byte
                        CertificateFactory cf = CertificateFactory.getInstance("X.509");
                        X509Certificate cert = (X509Certificate)cf.generateCertificate(is);
                        cert.checkValidity();
                        CertificateLogical cl = new CertificateLogical(cert);
                        is.close();
                        if(cl!=null)
                        {
                            myCredentials.put(cl.getHolder()+","+cl.getIssuer(), credentialsDir[y].getPath());
                            CredentialsValidity.put(cl.getHolder()+","+cl.getIssuer(),String.valueOf(cert.getNotAfter().getTime()));
                            System.out.println("IACCESS: Loaded X509IDCert: "+cl.getHolder()+","+cl.getIssuer());
                        }
                    }
                    catch(Exception e)
                    {
                        try
                        {
                            FileInputStream is = new FileInputStream(credentialsDir[y]); //for files that contain byte
                            byte[] buffer = new byte[(int)credentialsDir[y].length()] ;
                            is.read(buffer);
                            certAsString = new String(buffer);
                            X509V2AttributeCertificate certificate = createX509AttrObject( decodeCertificate( certAsString ));
                            certificate.checkValidity();
                            CertificateLogical cl = new  CertificateLogical( certificate );
                            is.close();
                            if(cl!=null)
                            {
                                myCredentials.put(cl.getHolder() + "," + cl.getAttribute() + "," + cl.getIssuer(), credentialsDir[y].getPath());
                                CredentialsValidity.put(cl.getHolder() + "," + cl.getAttribute() + "," + cl.getIssuer(),String.valueOf(certificate.getNotAfter().getTime()));
                                System.out.println("IACCESS: Loaded X509AttrCert: " + cl.getHolder() + "," + cl.getAttribute() + "," + cl.getIssuer());
                            }
                        }
                        catch(Exception ee)
                        {
                            System.err.println( "IACCESS: EXCEPTION: CertificateManager.loadMyCredentials: " + ee.toString());
                        }
                    }
                }
            }            
        } catch (Exception e) {
            System.err.println( "IACCESS: EXCEPTION: CertificateManager.loadMyCredentials: " + e.toString());
        }
        
    }
  
    /**Loads the private key in the memory and extracts the holder of the X.509 ID certificate in the 
     *corresponding to the private key in the SSL keystore. The holder is transformed into a logical 
     *representation. The function retrieves the first key entry in the store. */
public static String loadSSLKeystoreAndGetHolderDNToLogic()
{
          String holder =null;
          try
            {
               KeyStore ts = KeyStore.getInstance("JKS");
               ts.load(new FileInputStream(CredentialsConfig.DEFAULT_KEYS), CredentialsConfig.DEFAULT_PASSWORD.toCharArray());            
               Enumeration enumK = ts.aliases();
               for (; enumK.hasMoreElements(); ) 
               {
                   String alias = (String)enumK.nextElement();
	  	     if(ts.isKeyEntry(alias))
                     {
                        SSLPrivateKey = ts.getKey(alias,CredentialsConfig.DEFAULT_PASSWORD.toCharArray());
                        java.security.cert.X509Certificate cert = (X509Certificate) ts.getCertificate(alias);
                        cert.checkValidity();
                        holder = cert.getSubjectX500Principal().getName().toLowerCase();
                        holder = holder.substring(holder.indexOf("cn=")+3);

                        if(holder.indexOf("=") > -1)
                        {
                            holder = holder.substring(0, holder.indexOf("="));
                            holder = holder.substring(0, holder.lastIndexOf(","));
                            holder = CertificateLogical.makeLogical(holder);
                        }
                        else
                            holder = CertificateLogical.makeLogical(holder); 
                        break;
                     }
               }
               
            }
            catch(Exception e)
            {
                System.err.println("IACCESS: EXCEPTION: SSLKeyStore Load X509Certificate: "+e);
            }
        
            if(holder==null || SSLPrivateKey==null)
                System.err.println("IACCESS: ERROR: SSLKeyStore: Unable to find an X509 Certificate in the SSL keystore.");
            return holder;
}
    
    
    /** Loads trusted SOAs' Public Keys. */   
 public static ArrayList loadSOAPublicKeys() {
        
        ArrayList keys = new ArrayList();
     try {
       String SOAKeys = CredentialsConfig.CaKeys; 
       String SOAKeysPassword = CredentialsConfig.CaKeysPassword;
       
       char[] CAKeysPassword = CredentialsConfig.DEFAULT_PASSWORD.toCharArray();
            KeyStore keystore = KeyStore.getInstance("JKS"); 
            FileInputStream isK = new FileInputStream(SOAKeys);
            keystore.load(isK, SOAKeysPassword.toCharArray());
            isK.close(); 
            Enumeration enumK = keystore.aliases();
            for (; enumK.hasMoreElements(); ) {
                String alias = (String)enumK.nextElement();
		if(keystore.isCertificateEntry(alias) || keystore.isKeyEntry(alias)){
                    java.security.cert.Certificate certK = keystore.getCertificate(alias);
                    keys.add(certK.getPublicKey());
                }
            }
        System.out.println("IACCESS: Loaded trusted SOAs public keys: " + keys.size());   
        } catch (Exception e) {        
            System.err.println( "IACCESS: EXCEPTION: KeysManager.loadSOAPublicKeys: " + e.toString());
        } 
        return keys;     
  }    
  

 /** Loads trusted CAs' Public Keys. */   
 public static ArrayList loadCAPublicKeys() {

        ArrayList keys = new ArrayList();
        try {
            String CAKeys = CredentialsConfig.DEFAULT_TRUSTSTORE;
            String CAKeysPassword = CredentialsConfig.DEFAULT_PASSWORD;

            KeyStore keystore = KeyStore.getInstance("JKS"); 
            keystore.load(new FileInputStream(CAKeys), CAKeysPassword.toCharArray());
            Enumeration enumK = keystore.aliases();
            for (; enumK.hasMoreElements(); ) {
                String alias = (String)enumK.nextElement();
		if(keystore.isCertificateEntry(alias) || keystore.isKeyEntry(alias)){
                    java.security.cert.Certificate certK = keystore.getCertificate(alias);
                    keys.add(certK.getPublicKey());
                }
            }
        System.out.println("IACCESS: Loaded trusted CAs public keys: " + keys.size());   
        } catch (Exception e) {        
            System.err.println( "IACCESS: EXCEPTION: KeysManager.loadCAPublicKeys:" + e.toString());
        } 
        return keys;     
  }    
 
 
 /**
  * Calls loadSOAPublicKeys(), loadCAPublicKeys(), and BouncyCastleProvider()
  * if not already loaded/set up.
  */
 public static void loadTrustedPublicKeys()
 {
      if ( SOAPublicKeys == null )
          SOAPublicKeys = loadSOAPublicKeys();      
      if ( CAPublicKeys == null )
          CAPublicKeys = loadCAPublicKeys(); 
      if (BC_provider == null)
      {
        BC_provider = new BouncyCastleProvider();
        Security.addProvider( BC_provider );   
      }
 }
 
 /**
  * Loads a certificate from a local file (certFilePath).
  */
 public static String setCertificateFromFile (String certFilePath) {
        String certAsString = null;
        try {
            File inputcertfile = new File(certFilePath);
            FileInputStream is = new FileInputStream(inputcertfile); //for files that contain byte
            byte[] buffer = new byte[(int)inputcertfile.length()] ;
            is.read(buffer);
            certAsString = new String(buffer);
            is.close();
        } catch (Exception e) {
            System.err.println( "IACCESS: EXCEPTION: X509_certificate.setCertificateFromFile:" + e.toString());
        }
        return certAsString;
    }
     
  public static byte[] decodeCertificate( String certAsString) { 
        byte[] result = null;
        try {
            result = new BASE64Decoder().decodeBuffer(certAsString);
        } catch (Exception e) {
            System.err.println( "IACCESS: EXCEPTION: X509_certificate.decodeCertificate: " + e.toString());
        }
        return result;
    }  
  
public static X509V2AttributeCertificate createX509AttrObject(byte[] certDecoded) { 
         X509V2AttributeCertificate result = null;
         try {            
            result = new X509V2AttributeCertificate(certDecoded);             
         } catch (Exception e) {
            System.err.println( "IACCESS: EXCEPTION: CertificateManager.createX509AttrObject: " + e.toString());
         }
         return  result;
 };
 
 public static boolean verifyX509AttrCert(X509V2AttributeCertificate certificate, ArrayList PublicKeysOfCA, long sessionValidNotAfter) {
        boolean result = false;
        try {
            certificate.checkValidity(new Date(sessionValidNotAfter));
            if ( (checkSignatureAttrCert(certificate, PublicKeysOfCA) > -1 ) )
                result = true;
        } catch (Exception e) {
            System.err.println( "IACCESS: EXCEPTION: verifyX509AttrCert: " + e.toString());
        }
        return result;
 }

  public static boolean verifyX509IDCert(java.security.cert.X509Certificate certificate, ArrayList PublicKeysOfCA, long sessionValidNotAfter) {
        boolean result = false;
        try {
            certificate.checkValidity(new Date(sessionValidNotAfter));
            for (int i = 0;  i < PublicKeysOfCA.size(); i++) 
            {
                try 
                {
                    certificate.verify((PublicKey) PublicKeysOfCA.get(i));
                    return true;                
                } 
                catch (Exception e) 
                {
                    if ( i == (PublicKeysOfCA.size() - 1) )
                        System.err.println( "IACCESS: verifyX509IDCert.checkSignature: " + e.toString());
                }
            }
        } catch (Exception e) {
            System.err.println( "IACCESS: EXCEPTION: CertificateManager.verifyX509IDCert: " + e.toString());
        }
        return result;
 }

 public static int checkSignatureAttrCert(X509V2AttributeCertificate certificate, ArrayList PublicKeysOfCA) {
        int result = -1;
        for (int i = 0;  i < PublicKeysOfCA.size(); i++) {
            try {
                if ( certificate != null)
                    certificate.verify((PublicKey) PublicKeysOfCA.get(i), "BC");
                return i;                
            } catch (Exception e) {
                if ( i == (PublicKeysOfCA.size() - 1) )
                    System.err.println( "IACCESS: EXCEPTION: checkSignatureAttrCert: " + e.toString());
            }
        }
        return result;
 }
      
  /** Maps a certificate received over a secure SSL connection (Base64 encoded) to a logic format. */
  public static CertificateLogical mappingToLogicalSSL(String certAsString, long sessionValidNotAfter) {
      CertificateLogical result = null;
      try
      {
          CertificateFactory cf = CertificateFactory.getInstance("X.509");
          X509Certificate cert = (X509Certificate)cf.generateCertificate( new ByteArrayInputStream(certAsString.getBytes()) );
          if ( ! verifyX509IDCert(cert, CAPublicKeys, sessionValidNotAfter) )
          {
              result = null;
          }
          else
          {
              result = new CertificateLogical(cert);
              String str = result.getHolder()+","+result.getIssuer();
              if (CredentialsValidity.contains(str))
              {
                 long ln1 = Long.parseLong((String) CredentialsValidity.get(str));
                 long ln2 = cert.getNotAfter().getTime();
                 if(ln2>ln1)
                    CredentialsValidity.put(str,String.valueOf(ln2));      
              }
              else
                  CredentialsValidity.put(str,String.valueOf(cert.getNotAfter().getTime())); 
          }
      }
      catch(Exception e)      
      {
              try 
              {
                  X509V2AttributeCertificate certificate = createX509AttrObject( decodeCertificate( certAsString ));
                  if ( ! verifyX509AttrCert(certificate, SOAPublicKeys, sessionValidNotAfter) )
                  {
                      result = null;
                  }
                  else
                  {
                      result = new  CertificateLogical( certificate );
                      String str = result.getHolder()+","+result.getAttribute()+","+result.getIssuer();
                      if (CredentialsValidity.contains(str))
                      {
                         long ln1 = Long.parseLong((String) CredentialsValidity.get(str));
                         long ln2 = certificate.getNotAfter().getTime();
                         if(ln2>ln1)
                            CredentialsValidity.put(str,String.valueOf(ln2));      
                      }
                      else
                          CredentialsValidity.put(str,String.valueOf(certificate.getNotAfter().getTime())); 
                  }
              }  
              catch (Exception ee) { 
                    System.err.println( "IACCESS: EXCEPTION: CertificateManager.mappingToLogicalSSL: " + ee.toString());
              } 
      }
      return result;
  }
  
    public static String getCredential(String cred, String anyValueTerm) {
        
        //Note that a Hashtable is a synchronized element
        try {

            String Request_holder =null;
            String Request_attr=null;
            String Request_issuer=null;
            String tomatch = null;
            //default (false) is for a credential text,text,text
            boolean isCertificate = false; 
            String[] tmpstr = Pattern.compile("\\,").split(cred);
            if(tmpstr.length == 2)//certificate text,text
            {
                isCertificate = true;
                Request_holder = tmpstr[0];
                Request_issuer = tmpstr[1];
                if(Request_holder.compareTo(anyValueTerm) == 0)
                    tomatch = Request_issuer;
            }
            else // credential text,text,text
            {
                Request_holder = tmpstr[0];
                Request_attr = tmpstr[1];
                Request_issuer = tmpstr[2];
                if(Request_holder.compareTo(anyValueTerm) == 0)
                    tomatch = Request_attr+","+Request_issuer;
            }
                     
            if(tomatch == null)
            {
                    if (myCredentials.containsKey(cred))
                        return new String(cred);
                    else
                        return null;
            }
            else // anyvalueterm is found 
            {

                Enumeration allEntities =  myCredentials.keys();
                for (; allEntities.hasMoreElements(); ) 
                {
                        String nextEntity = (String)allEntities.nextElement(); 
                        if(((String)nextEntity).indexOf(tomatch)>-1 &&
                           ((String)nextEntity).indexOf(anyValueTerm)==-1 )
                        {
                            if(!isCertificate)
                                return new String((String)nextEntity);
                            else
                            {  //check if the matched one is also a certificate
                                int fi = ((String)nextEntity).indexOf(',');
                                int li = ((String)nextEntity).lastIndexOf(','); 
                                if((fi>-1) && (li>-1) && (fi == li))
                                    return new String((String)nextEntity);
                            }
                        }
                }
            }
        } catch (Exception e) 
        { 
            System.err.println("IACCESS: EXCEPTION: CredentialManager.getCredential: "+ e);
        }         
        //if this point no cred found.
        return null;
    }
  
public static String getCertFilePath(String ServiceRequest)
{ 
    String toReturn = null; 
    try{
        toReturn = (String) myCredentials.get(ServiceRequest);
    }catch(Exception e)
    {
        System.err.println("IACCESS: EXCEPTION: CredentialManager.getCertFilePath: "+ e);
    }
    return toReturn;
}
    
    
/** Generates a SAML Authorization Decision Statement. */
public static String accessDecisionToSAML(int decision, 
                                          String servReq,
                                          String sessionID,
                                          String Identity,
                                          long accessDecisionTimeValidity)
  {
    try
      {
      // Initialize the xml security library
      if (!org.apache.xml.security.Init.isInitialized())
         org.apache.xml.security.Init.init();
      
      String[] tmpstr = Pattern.compile("\\,").split(servReq);                            
      String ServiceRequest = tmpstr[0]; 
      String ServiceAction = tmpstr[1]; 
      
      SAMLAuthorizationDecisionStatement samlADS = new SAMLAuthorizationDecisionStatement();
      //set access decision, service request and service action
      if(decision==1)
          samlADS.setDecision(SAMLDecision.PERMIT);
      else
          samlADS.setDecision(SAMLDecision.DENY);
      samlADS.setResource(ServiceRequest);
      //set service action 
      SAMLAction samlAc = new SAMLAction();
      samlAc.setData(ServiceAction);
      Vector v = new Vector();
      v.addElement(samlAc);
      samlADS.setActions(v);

      //set subject name. NB the subject name is the 
      //serial number of the SSL certificate. 
      SAMLSubject samlSubj = new SAMLSubject();
      SAMLNameIdentifier samlNId = new SAMLNameIdentifier();
      samlNId.setName(Identity);
      samlNId.setFormat(samlNId.FORMAT_X509);
      samlSubj.setNameIdentifier(samlNId);
      samlADS.setSubject(samlSubj);
            
      //new SAML Assertion that includes the SAML Authz Decision Statement
      //set SessionID and IACCESS Issuer
      SAMLAssertion samlAs = new SAMLAssertion();
      samlAs.setId(sessionID);
      samlAs.setIssuer("IACCESS_AUTHORIZATION_SERVICE");
      
      //set decision time validity 
      long ct = System.currentTimeMillis();
      samlAs.setNotBefore(new Date(ct));
      samlAs.setNotOnOrAfter(new Date(ct+accessDecisionTimeValidity));
      
      //add the SAML Authz Decision Statement in the new SAML Assertion
      Vector vv = new Vector();
      vv.addElement(samlADS);
      samlAs.setStatements(vv);
      
      if(SSLPrivateKey!=null)
          samlAs.sign(XMLSignature.ALGO_ID_SIGNATURE_RSA_SHA1,SSLPrivateKey,null);
      else
          System.err.println("IACCESS: ERROR: No SSL Private Key found to sign a SAML Assertion.");
          
      //return samlAs.toBase64(); 
      return new sun.misc.BASE64Encoder().encode(samlAs.toString().getBytes());
            
        }
        catch(Exception e)
        {
            System.err.println("IACCESS: EXCEPTION: CertificateManager.accessDecisionToSAML: " + e.toString());
            return null;
        }
 }
}
