/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.TNProtocol;

import java.io.*;
import java.net.*;
import java.util.Date;
import java.util.Calendar;
import java.util.logging.Level;
import java.util.Vector;
import java.util.regex.Pattern;
import javax.net.ssl.*;
import java.security.KeyStore;
import java.security.cert.*;


import org.iaccess.CredentialManager.*;
/**
 * Simutales a JBOSS server. It uses TCP/IP sockets to send/receive information.
 * Should be deprecated in future releases and implemented with another application server, like Tomcat, Glassfish etc.
 * <p>More details are given to each of the functions in the class.
 * @see org.iaccess.CredentialManager.CredentialsConfig
 * @see org.iaccess.TNProtocol.iAccessDispatcher
 */
public class JBOSS extends Thread{

        /** Specifies the maximum queue length for incoming connection indications
        * (a request to connect) is set to 500 by default. If a connection 
        * indication arrives when the queue is full, the connection is refused.*/
	public static int queueLength = 500;
        /** Specifies the maximum lenght of message in bytes.*/
	public static int RequestLength = 256*4;
  
        /** Specifies the link to iAccessDispatcher. A JBOSS object notifies AccessDispatcher when a new request appears.*/
	private iAccessDispatcher AccessDispatcher;
        /** Indicates the current server's state. By default stopWorking = false.*/
	private volatile boolean stopWorking = false;

        //SSL add
        private static final String DEFAULT_PASSWORD= CredentialsConfig.DEFAULT_PASSWORD;
        private static final String DEFAULT_KEYS= CredentialsConfig.DEFAULT_KEYS;
        private static final String DEFAULT_TRUSTSTORE= CredentialsConfig.DEFAULT_TRUSTSTORE;
        
        private static char[] passphrase = DEFAULT_PASSWORD.toCharArray();
        private static String keys = DEFAULT_KEYS;
        private static String truststore = DEFAULT_TRUSTSTORE;        
  
        private static SSLContext ctx = getSSLContext();
        
        protected static SSLContext getSSLContext() {
	    SSLContext ctx = null;
	    try {
		ctx = SSLContext.getInstance("TLS");
		
        KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
		KeyStore ks = KeyStore.getInstance("JKS");
		ks.load(new FileInputStream(keys), passphrase);
		kmf.init(ks, passphrase);
		
		TrustManagerFactory tmf = TrustManagerFactory.getInstance("SunX509");
		KeyStore ts = KeyStore.getInstance("JKS");
		ts.load(new FileInputStream(truststore), passphrase);
		tmf.init(ts);
		
		ctx.init(kmf.getKeyManagers(), tmf.getTrustManagers(), null);
                //System.err.println("SSLContext is initialized...");      
                System.out.println("IACCESS: SSL handshake data has been successfully loaded.");
	    } catch (Exception e) {
		System.err.println("IACCESS: EXCEPTION: JBOSS.getSSLContex: " + e.getMessage());
	    }  
            
            return ctx;
        }
        //SSL add
        
        
        /** Creates a new instance of the JBOSS class. Class java.util.Thread is inherited by JBOSS, i.e. JBOSS is always run in a new thread. */
	public JBOSS(iAccessDispatcher AccessDispatcher) {
		super ("JBOSS_Simulator");
		this.AccessDispatcher  = AccessDispatcher;
	}
	
 
	public void run(){		
	try{	
            SSLServerSocketFactory ssf = ctx.getServerSocketFactory();
            SSLServerSocket serverSocket=(SSLServerSocket) ssf.createServerSocket(TSessionInfo.TCPportIn);
            serverSocket.setNeedClientAuth(CredentialsConfig.needClientAuth);
            //serverSocket.
            byte[] Buffer = new byte[RequestLength];
            TRequest Request;             
            while (true) {			
                    SSLSocket SocketWaitingResult = (SSLSocket)serverSocket.accept();                
                        
                    SocketWaitingResult.addHandshakeCompletedListener(
                        new HandshakeCompletedListener() {
                        public void handshakeCompleted( HandshakeCompletedEvent event) {
                            try {
                                //System.out.println("IACCESS: SSL handshake is completed!" + CredentialsConfig.needClientAuth);
                                //update user profile here by certificate presented during SSL handshake
                                if ( event.getPeerCertificates().length > 0) {
                                    java.security.cert.X509Certificate cert = ((java.security.cert.X509Certificate) event.getPeerCertificates()[0]);
                                    //Check if the SSL certificate from the handshake has validity current time+session time + dynamic session extention
                                    cert.checkValidity(new Date(System.currentTimeMillis()+TSessionInfo.SessionTimeLimit+TSessionInfo.DynamicSessionTimeExtension));
                                    CertificateLogical certLog = new  CertificateLogical(cert);                          
                                    if(certLog==null)
                                        throw new Exception("Unable to extract data from the SSL certificate.");

                                    //Update Cert X500Subject linking to the SSL certificate X500Subject.CN logical representation 
                                    UserProfile.updateOngoingProfile( cert.getSubjectX500Principal().getName(), certLog.getHolder()+","+certLog.getIssuer());  
                                    //Update holder logical value linking to the SSL certificate logical representation
                                    UserProfile.updateOngoingProfile( certLog.getHolder(), certLog.getHolder()+","+certLog.getIssuer());  
                                    //Add SSL cert validity in the hashtable of certificate manager
                                    String str = certLog.getHolder()+","+certLog.getIssuer();
                                      if (CertificateManager.CredentialsValidity.contains(str))
                                      {
                                         long ln1 = Long.parseLong((String) CertificateManager.CredentialsValidity.get(str));
                                         long ln2 = cert.getNotAfter().getTime();
                                         if(ln2>ln1)
                                            CertificateManager.CredentialsValidity.put(str,String.valueOf(ln2));      
                                      }
                                      else
                                          CertificateManager.CredentialsValidity.put(str,String.valueOf(cert.getNotAfter().getTime())); 
                                }
                                
                            }
                            catch(CertificateExpiredException ce)
                            {
                                System.err.println("IACCESS: EXCEPTION: JBOSS.handshakeCompleted:SSL Certificate not valid for a session duration (including dynamic extension): " + ce.toString());
                            }
                            catch (Exception e) {
                                System.err.println("IACCESS: EXCEPTION: JBOSS.handshakeCompleted: " + e.toString());
                            }
                        }
                        }
                    );  
                    try{ 
			if (stopWorking) {
                            SocketWaitingResult.close();
                            serverSocket.close();
                            break;
			}
			InputStream inStream = SocketWaitingResult.getInputStream();
			inStream.read(Buffer);								
			Request = GetTRequest (Buffer);
                        System.out.println("IACCESS: Received Request over SSL: " + Request.ServiceRequest);  
			AccessDispatcher.receiveRequest(Request, SocketWaitingResult);
                    } catch (Exception e) {
                        System.err.println("IACCESS: EXCEPTION: JBOSS.run"); e.printStackTrace();
			iAccessDispatcher.logger.logp(Level.SEVERE, "JBOSS", "runSSL", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
                    }
		}
        } catch (Exception e) {
            System.err.println("IACCESS: EXCEPTION: JBOSS.run"); e.printStackTrace();
            iAccessDispatcher.logger.logp(Level.SEVERE, "JBOSS", "runSSL", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
        }
 
}
        
     /** Sends response to the request and closes the connection.
         * TCPportOut defined at TSessioninfo is used for the connection.
         * @param Result grant(1) or deny(0).
         * @param s Denotes the socket awaiting for result.
         * @see org.iaccess.TNProtocol.TSessionInfo
         * @exception Exception is thrown when the remote connection fails.
         */
        public static void sendResponseSSL (TRequest Request, int Result, SSLSocket s) throws Exception{

            String toSend = "deny";
            String samlAssertion = null;
            String certFilePath = null;
            try
            {
                                
            //if grant and request for a credential
            if ( (Result == iAccessNegotiation.grant) && (!Request.IsService) ) 
            {
                certFilePath = CertificateManager.getCertFilePath(Request.ServiceRequest);
                if ( certFilePath == null) {
                    System.err.println("IACCESS: ERROR SEVERE: certificates and policies are not consistent: " + Request.ServiceRequest);
                    Result = iAccessNegotiation.deny;
                } 
                else
                    toSend = CertificateManager.setCertificateFromFile(  certFilePath );                 
            } 
            else 
            if(Request.IsService)
            {  // Response for a service request
                if ( Result == iAccessNegotiation.grant )
                {
                   toSend = "grant";
                   //Get Suject X500 Name of the ID certificate from the SSL channel 
                   String X500Name = UserProfile.getX500Name(Request.Identity);
                   if(X500Name!=null)
                   {
                        samlAssertion = CertificateManager.accessDecisionToSAML(Result, 
                                                                        Request.ServiceRequest, 
                                                                        Request.SessionID, 
                                                                        X500Name,
                                                                        ((TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).AccessDecisionTimeValidity
                                                                        );
                        if(samlAssertion!=null)
                        {
                           toSend += samlAssertion;
                           //Store the SAML token in a file in the log directory.
                           FileOutputStream fout = new FileOutputStream(new File(System.getProperty("user.dir")+TSessionInfo.ActualFileSeparator+"log"+TSessionInfo.ActualFileSeparator+Request.SessionID+"_SAMLToken.xml"));
                           fout.write(new String(CertificateManager.decodeCertificate(samlAssertion)).getBytes());
                           fout.close();
                           
                        }
                   }
                }
                else
                   toSend = "deny";
                
            }

            System.out.println( "IACCESS: Sending Response over SSL for: " + Request.ServiceRequest + ":"  + Result);
            System.out.println("IACCESS: ---------: begin :----------");
            System.out.println(toSend);
           
            System.out.println("IACCESS: ---------: end :----------");
            s.getOutputStream().write(toSend.getBytes());
            s.close();
            }
            catch(Exception e)
            {
                System.err.println("IACCESS: EXCEPTION: JBOSS.sendResponseSSL: " + e);
                if(s!=null) s.close();
            }
	}
        
        /** Sends the counter-request and awaits for the result.
         * @param Request Specifies the request which  should be sent.
         * @param inetAddress All requests are sent to this address to TCPportOut defined at TSessioninfo.
         * @return the  result of request. It may be grant(1) or deny(0).
         * @see org.iaccess.TNProtocol.TSessionInfo
         * @exception Exception is thrown when it is not possible to establish a remote connection.
         */
 	public static String sendRequestSSL(TRequest Request, InetAddress inetAddress) throws Exception {
            SSLSocket s = null;
            try{
                    TRequest newRequest = null;
                    SSLSocketFactory sf =ctx.getSocketFactory();
                    s = (SSLSocket)sf.createSocket( inetAddress, ((TSessionInfo)iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).TCPportOut );
                    
                    String anyValueTerm = TSessionInfo.CredAnyValueTerm;
                    String anyValueTermOpponent = ( (TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).GetCredAnyValueTermOpponent();

                    //Substitu the opponent 
                    if(!Request.IsService && Request.ServiceRequest.startsWith(anyValueTerm))
                    {
                        newRequest = new TRequest(Request);
                        newRequest.ServiceRequest = anyValueTermOpponent + newRequest.ServiceRequest.substring(newRequest.ServiceRequest.indexOf(","));
                    }
                    else
                    {
                        newRequest = Request;
                    }

                    System.out.println("IACCESS: Sending Request over SSL: "  + newRequest.ServiceRequest);

                    long validNotAfter = ( (TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).GetValidNotAfter();
                    long negStartTime = ( (TSessionInfo) iAccessDispatcher.SessionInfoTable.get(Request.SessionID)).negStartTime;
                    long dynamicSessionTimeExt = TSessionInfo.DynamicSessionTimeExtension;
                    long timeToWait = 0;
                    if(Request.IsService)
                        timeToWait = (validNotAfter + dynamicSessionTimeExt) - Calendar.getInstance().getTimeInMillis();
                    s.setSoTimeout((int) timeToWait);
                    s.getOutputStream().write(JBOSS.GetRawRequest(newRequest));

                    //read the whole certificate from the secure channel!
                    String certificateAsString = new String("");
                    BufferedReader inFromClient = new BufferedReader (new InputStreamReader(s.getInputStream()));
                    String inputLine= inFromClient.readLine(); 

                    while (inputLine != null) {
                        certificateAsString += inputLine+"\n";
                        inputLine = inFromClient.readLine();                 
                    }
                    s.close(); 


                    System.out.println( "IACCESS: Received Response over SSL for: " + newRequest.ServiceRequest) ;
                    System.out.println("IACCESS: ---------: begin :----------");
                    System.out.println( certificateAsString );
                    System.out.println("IACCESS: ---------: end :----------");    

                    //return null if deny 
                    String result = null;
                    if ( ! (certificateAsString.startsWith("deny") || certificateAsString.startsWith("grant")) ) 
                    {
                        //check if request for a service returns response different from grant/deny
                        if(Request.IsService)
                        {
                            System.err.println("IACCESS: ERROR: SendRequestSSL: Unsupported protocol format for a message response of a service request. Response must begin with \"grant\" or \"deny\".");
                            return null;
                        }
                        
                        //result is null if distrusted credential
                        CertificateLogical tempCertLog;
                        if(negStartTime+TSessionInfo.SessionTimeLimit < validNotAfter)
                           tempCertLog = CertificateManager.mappingToLogicalSSL( certificateAsString, validNotAfter );
                        else
                           tempCertLog = CertificateManager.mappingToLogicalSSL( certificateAsString, validNotAfter+dynamicSessionTimeExt );
                        
                        if(tempCertLog != null)
                        {
                            int fi = Request.ServiceRequest.indexOf(',');
                            int li = Request.ServiceRequest.lastIndexOf(','); 
                            String[] tmpstr = Pattern.compile("\\,").split(Request.ServiceRequest);
                            if(fi<li)  //check if Attribute Certificate has been requested
                            {
                                //CHECK IF IDENTITY CERT HAS BEEN GIVEN
                                //return null
                                if(tempCertLog.getAttribute()==null)
                                {
                                    System.err.println("IACCESS: EXCEPTION: SendRequestSSL: Requested Attribute credential BUT received ID credential: " + tempCertLog.getHolder()+","+tempCertLog.getIssuer());
                                    return null;
                                }
                                String Request_holder = tmpstr[0];
                                String Request_attr = tmpstr[1];
                                String Request_issuer = tmpstr[2];

                                if(Request_holder.compareTo(anyValueTerm)!=0)
                                {
                                    if((Request_holder.compareTo(tempCertLog.getHolder())==0) &&
                                       (Request_attr.compareTo(tempCertLog.getAttribute())==0) &&
                                       (Request_issuer.compareTo(tempCertLog.getIssuer())==0))
                                    {
                                       result = "grant";    
                                    }
                                    else
                                    {
                                        System.err.println("IACCESS: EXCEPTION: SendRequestSSL: Requested credential "+Request_holder+","+Request_attr+","+Request_issuer+" BUT received " + tempCertLog.getHolder() + "," + tempCertLog.getAttribute() + "," + tempCertLog.getIssuer());
                                    }
                                }
                                else //If holder has anyvalue term specified. 
                                {
                                    if((Request_attr.compareTo(tempCertLog.getAttribute())==0) &&
                                       (Request_issuer.compareTo(tempCertLog.getIssuer())==0))
                                    {
                                        // we return "grantholdername,attribute,issuer" so that we extract it in iAccessNegotiation
                                        result = "grant"+tempCertLog.getHolder() + "," + tempCertLog.getAttribute() + "," + tempCertLog.getIssuer();    
                                    }
                                    else
                                    {
                                        System.err.println("IACCESS: EXCEPTION: SendRequestSSL: Requested credential "+Request_holder+","+Request_attr+","+Request_issuer+" BUT received " + tempCertLog.getHolder() + "," + tempCertLog.getAttribute() + "," + tempCertLog.getIssuer());
                                    }
                                }
                            }
                            else
                            {
                               // here comes the check for
                               // certificates and if they match as requested
                                if(tempCertLog.getAttribute()!=null)
                                {
                                    System.err.println("IACCESS: EXCEPTION: SendRequestSSL: Requested ID certificate BUT received Attribute credential: " + tempCertLog.getHolder()+","+tempCertLog.getAttribute()+","+tempCertLog.getIssuer());
                                    return null;
                                }
                                String Request_holder = tmpstr[0];
                                String Request_issuer = tmpstr[1];
                                if(Request_holder.compareTo(anyValueTerm)!=0)
                                {
                                    if((Request_holder.compareTo(tempCertLog.getHolder())==0) &&
                                       (Request_issuer.compareTo(tempCertLog.getIssuer())==0))
                                    {
                                       result = "grant";    
                                    }
                                    else
                                    {
                                        System.err.println("IACCESS: EXCEPTION: SendRequestSSL: Requested credential "+Request_holder+","+Request_issuer+" BUT received " + tempCertLog.getHolder() + "," + tempCertLog.getIssuer());
                                    }
                                }
                                else //If holder has anyvalue term specified. 
                                {
                                    if((Request_issuer.compareTo(tempCertLog.getIssuer())==0))
                                    {
                                        // we return "grantholdername,attribute,issuer" so that we extract it in iAccessNegotiation
                                        result = "grant"+tempCertLog.getHolder() + "," + tempCertLog.getIssuer();    
                                    }
                                    else
                                    {
                                        System.err.println("IACCESS: EXCEPTION: SendRequestSSL: Requested credential "+Request_holder+","+Request_issuer+" BUT received " + tempCertLog.getHolder() + "," + tempCertLog.getIssuer());
                                    }
                                }
                            }
                        }
                        else
                            System.err.println("IACCESS: EXCEPTION: Received expired or distrusted credential.");
                    }
                    else
                        if ( certificateAsString.startsWith("grant")) 
                        {
                            result = "grant";
                            if(certificateAsString.length()>"grant".length())
                            {
                                //HERE WE SHOULD CHECK (TO BE DONE) IF THE RECEIVED CERTIFICATE IS SIGNED BY THE OPPONENT
                                //AND ATTESTING THE GRANT DECISION TO THE REQUESTED RESOURCE AND ACTION.
                                String samlAssertion =  certificateAsString.substring("grant".length());
                                // actually no need of explicit grant in the protocol message if the check above is done.
                                result = new String(CertificateManager.decodeCertificate(samlAssertion));
                            }
                        }
                    
                    //return null if really deny or received certificated has already expired!
                    return result;
                }
                catch(Exception ex)
                {
                    System.err.println("IACCESS: EXCEPTION: JBOSS.sendRequestSSL: " + ex.toString());
                    if(s!=null) s.close();
                    return null;
                }
                
	}        
	
        /** Transforms the received raw request (array of bytes of <b>RequestLength</b> elements) to hight level TNProtocol.TRequest class.
         * After this transformation the request is redirected to the iAccessDispatcher.
         * @see org.iaccess.TNProtocol.TRequest
         */
	public static TRequest GetTRequest(byte[] RawRequest) {
		try {
			int i = 1;
			TRequest Request = new TRequest();
			Vector Credentials = new Vector(); 
			
//			IsService
			if ( (char)RawRequest[0] == '1')
				Request.IsService = true;
			else 
				Request.IsService = false;
//			SessionID
			do {
				Request.SessionID += (char)RawRequest[i]; i++;						
			} while ( RawRequest[i] != 11);
//			Definition
			while( RawRequest[i] == 11)	i++;
			do {
				Request.ServiceRequest += (char)RawRequest[i];	i++;										
			} while (RawRequest[i] != 11);
//                      Identity                        
			while( RawRequest[i] == 11)	i++;
			do {
				Request.Identity += (char)RawRequest[i];	i++;										
			} while (RawRequest[i] != 11);                        
//			Credentials			
			while(RawRequest[i] == 11)	i++;
			
			int j = 0;
			String CurrCredential; 
			while (RawRequest[i] != -1) {				
				CurrCredential = new String();
				do {
					CurrCredential += (char)RawRequest[i];	i++;
					if (RawRequest[i] == -1) break;
				} while ((RawRequest[i]!= 11));
				while(RawRequest[i] == 11)	i++;
				j++;
				Credentials.addElement(CurrCredential);
				CurrCredential = new String();
			}		
			
			Request.Credentials = new String[Credentials.size()];
			for (int ii = 0; ii < Credentials.size();ii++) {
				Request.Credentials[ii] = (String) Credentials.get(ii);	
			}			
			return Request;
		}catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "JBOSS", "GetRequest", "Thread: "+ Thread.currentThread().getName() +"; Error: "+ e.toString());
                        
			return null;
		}		
	}

        /** Transforms the hight level request (TNProtocol.TRequest class)
         * to the the raw request (array of bytes of <b>RequestLength</b> elements) wchich can be sent over TCP/IP.
         * @see org.iaccess.TNProtocol.TRequest
         */
	public static byte[] GetRawRequest(TRequest Request) {
		try {
			String RawRequest = new String();
//			IsService
			if (Request.IsService)
				RawRequest += "1";
			else
				RawRequest += "0";
//			SessionID
			RawRequest += Request.SessionID;
			RawRequest += (char) 11;
//			Definition
			RawRequest += Request.ServiceRequest;
			RawRequest += (char) 11;
//                      Identity
			RawRequest += Request.Identity;
			RawRequest += (char) 11;                        
//			Credentials			
			for (int i = 0; i < Request.Credentials.length; i++) {
				RawRequest += Request.Credentials[i];
				RawRequest += (char) 11;				 
			}
			RawRequest += (char) 11;

			byte[] buffer = RawRequest.getBytes();
			buffer[RawRequest.length()-1] = -1;
			
			return buffer;
		}catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "JBOSS", "GetRawRequest", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
                       
			return null;
		}		
	}	
	
        /** Sends the stopWorking message to JBOSS server.
         * No any more messages will be accepted by the server.
         */
	public void Stop() {
		try {
			stopWorking = true;
			Socket stopRequest = new Socket(InetAddress.getLocalHost().getHostName(),TSessionInfo.TCPportIn);
			stopRequest.close();
		} catch (Exception e) {
			iAccessDispatcher.logger.logp(Level.SEVERE, "JBOSS", "Stop", "Thread: "+ Thread.currentThread().getName() +"; EXCEPTION: "+ e.toString());
                        
		}
	}

}