/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.TNProtocol;

/**
 * Defines the high level request structure used for negotiations.
 * Any credential/service request is of type TRequest.
 */
public class TRequest {
	/** SessionID = String(CurrentTime) + String (Random). */
	public String SessionID;        
        /** If Request for the credential/certificate this field is false. */
	public boolean IsService;        
        /** Service request definition. */
	public String ServiceRequest;        
        /**
         * Defines additional negotiation information.
         * It was originally designed for an initial set of cderentials/certificates.
         * Since we disable an input set of credentials for a service/credential request
         * we re-use this field for transmitting additional information of a negotiation process,
         * such as SystemAnyValue term (Credentials[0]), and TCPPortIn for incoming responses/counter requests to a current request (Credentials[1]).
         * One can use it for adding more negotiation information to be sent to an opponent, such as session time, type of security tokens (SAML, SPKI, X.509), etc.
         */
	public String[]	Credentials;	
        /**Identity of an entity sending the request. */
    public String Identity;
        
        /**
         * Creates a new instance of TRequest.
         * All fieldls are initialized as empty objects, IsService = "false".
         */
	public TRequest() {
		SessionID = new String();
		ServiceRequest = new String();
                Identity = new String();
		Credentials = new String[]{};
		IsService = false;
	}
	
        /**
         * A copying constructor.
         */
	public TRequest (TRequest Request){
            
       		if (Request.SessionID !=  null)
                    SessionID = new String(Request.SessionID);
                else
                    SessionID = null;
		ServiceRequest = new String(Request.ServiceRequest);
                Identity = new String(Request.Identity);
		if ( Request.Credentials != null)
			Credentials = (String[]) Request.Credentials.clone();
		IsService = Request.IsService;
	}
        
	public TRequest(String SessionID, boolean IsService,String ServiceRequest, String Identity){
		if (SessionID !=  null)
			this.SessionID = new String(SessionID);
		else
			this.SessionID = null; 
		this.IsService = IsService;
		this.ServiceRequest = new String(ServiceRequest);
                this.Identity = new String(Identity);
		this.Credentials = new String[]{""};
	}	


	public TRequest(String SessionID, boolean IsService,String ServiceRequest, String Identity, String SystemAnyValueTerm){
		if (SessionID !=  null)
			this.SessionID = new String(SessionID);
		else
			this.SessionID = null; 
		this.IsService = IsService;
		this.ServiceRequest = new String(ServiceRequest);
                this.Identity = new String(Identity);
		this.Credentials = new String[]{SystemAnyValueTerm};
	}	

        /**TCPportIn is actually defining the listening port for this request. 
         * By default TCPportIn is the second argument of the optional String[] Credentials of the request. */
	public TRequest(String SessionID, boolean IsService,String ServiceRequest, String Identity, String SystemAnyValueTerm, int TCPportIn){
		if (SessionID !=  null)
			this.SessionID = new String(SessionID);
		else
			this.SessionID = null; 
		this.IsService = IsService;
		this.ServiceRequest = new String(ServiceRequest);
        this.Identity = new String(Identity);
		this.Credentials = new String[]{SystemAnyValueTerm,Integer.toString(TCPportIn)};
	}	
        
        
}