/*
 * Copyright 2006, 2008 Hristo Koshutanski, Aliaksandr Lazouski, Associazione CREATE-NET, and Universita' degli Studi di Trento.  
 * 
 * This file is part of iAccess authorization service software (called iAccess).
 * iAccess is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * iAccess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with iAccess.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.iaccess.UserInterface;

import java.io.*;
import java.util.regex.Pattern;
import java.net.InetAddress;
import org.iaccess.TNProtocol.TSessionInfo;
import org.iaccess.TNProtocol.iAccessDispatcher;

/** Generates an iAccess agent GUI. It runs the iAccessDispatcher and then
 * visualizes the interface components. It has a nested class OUTputStream that 
 * visualizes all relevant negotiation messages. 
 * @see org.iaccess.TNProtocol.iAccessDispatcher
 * @see org.iaccess.TNProtocol.TSessionInfo
 */
public class UserInterface extends javax.swing.JFrame {
    private iAccessDispatcher TestDispatcher;
    private java.awt.TextField Edit_DSTimeExtension;
    private java.awt.TextField Edit_STimeLimit;
    private java.awt.TextField Edit_ServerAddress;
    private java.awt.TextField Edit_ServerPort;
    private java.awt.TextField Edit_ServiceAction;
    private java.awt.TextField Edit_ServiceName;
    private java.awt.TextArea TextArea_ClientWorkSpace;
    private java.awt.Button btn_StartNegotiations;
    private java.awt.Label label1;
    private java.awt.Label label2;
    private java.awt.Label label3;
    private java.awt.Label _label5;
    private java.awt.Label label6;// for ":" of the port text field
    private java.awt.Label label7;
    private java.awt.Label _label8;
    private java.awt.TextArea TextArea_ServerWorkSpace;
    private java.awt.Label labelServer;
    private javax.swing.ImageIcon iaccess_logo;
    private String DisplayableSessionID = new String("0");
    private String logDataPathStr = new String("");
    private OUTputStream OutputDataStream = new OUTputStream();

    static public String UserName = new String("");        
   
    static UserInterface MyInterfaceLink = null;
       
    private class OUTputStream extends OutputStream {	

		public void write(int j){}
                
		public void write(byte[] Array) {
			synchronized (this) {
				String TempString = new String("");
				String TempString1 = new String("");
				String TempString2 = new String("");
                                String tab = "     ";
                                boolean flag = false;
				if ((Array[2])== '>') {TempString = tab; flag=true;}
				if ((Array[2])== '<') {TempString = tab; flag=true;}
		
                                int i = 0;
				for ( i = 0; i < Array.length; i++) {
					if (Array[i] == '!'){
                                                TempString2 = TempString;
                                                TempString = "";
						continue;
					}
                                        // do not display sesssion id 
                                        if (Array[i] == '}'){
                                            if(flag)
                                            {
                                               String[] st = Pattern.compile(":").split(TempString);
                                               TempString = TempString1 + st[0]  + " :: " +"<"+TempString2+">" + "............" +st[1];
                                            }
                                            break;                                         
                                        }    
                                        
					TempString += (char) Array[i];
					if (i == 2 && flag)
                                        {
						TempString1 = TempString + tab;
                                                TempString = "";
                                        }
				}
				
                                //get SessionID
                                String SessionID = new String("");
                                for (int j = i+1; j < Array.length; j++) {
                                   SessionID += (char) Array[j]; 
                                }
                                
                                // only one session info (client side) is displayable
                                if ( DisplayableSessionID.equals("0") && ( TempString.indexOf("outgoing service request") != -1) )
                                {
                                    DisplayableSessionID = SessionID;
                                    TextArea_ClientWorkSpace.setText(TextArea_ClientWorkSpace.getText() + "*** NEW SESSION <"+SessionID+"> ***" + "\n"); 
                                }
                                if (DisplayableSessionID.equals(SessionID))
                                    TextArea_ClientWorkSpace.setText(TextArea_ClientWorkSpace.getText() + TempString + "\n"); 
                                else 
                                    TextArea_ServerWorkSpace.setText(TextArea_ServerWorkSpace.getText() + TempString + "\n");
                                
                                //Update correctry session logs
                                File logDataPath = new File("log");
                        
                                if ( !logDataPath.exists() && !logDataPath.mkdir() ) 
                                        logDataPathStr = "";
                                else
                                    logDataPathStr = "log"+TSessionInfo.ActualFileSeparator;
                                
                                FileOutputStream out; 
                                PrintStream p;
     
                                try {
                                    out = new FileOutputStream( logDataPathStr +  SessionID + ".log", true);
                                    p = new PrintStream( out );
                                    p.println (TempString );
                                    p.close();
                              }  catch (Exception e){
                                    System.err.println ("IACCESS: EXCEPTION: Writing to a log file: " + e);
                              }                                     
			}		
		}
	}	

    /** Creates a GUI of an iAccess agent. */
    public UserInterface() {
        initComponents();
    }
    
    /** A method called from within the constructor to initialize the form.
     * The content of the method is (always) regenerated by the Form Editor.
     */
    private void initComponents() { 

        java.awt.Dimension screenSize = java.awt.Toolkit.getDefaultToolkit().getScreenSize();
    	
        btn_StartNegotiations = new java.awt.Button();
        label1 = new java.awt.Label();
        label2 = new java.awt.Label();
        label3 = new java.awt.Label();
        Edit_STimeLimit = new java.awt.TextField();
        Edit_DSTimeExtension = new java.awt.TextField();
        TextArea_ClientWorkSpace = new java.awt.TextArea();
        _label5 = new java.awt.Label();
        label6  = new java.awt.Label();
        label7  = new java.awt.Label();
        _label8 = new java.awt.Label();
        Edit_ServerAddress = new java.awt.TextField();
        Edit_ServerPort = new java.awt.TextField();
        Edit_ServiceAction = new java.awt.TextField();
        Edit_ServiceName = new java.awt.TextField();
        
        TextArea_ServerWorkSpace = new java.awt.TextArea();
        labelServer = new java.awt.Label();
        iaccess_logo = new javax.swing.ImageIcon("data/iaccess_logo.png");
        javax.swing.JLabel logo_background = new javax.swing.JLabel(iaccess_logo);

        getContentPane().setLayout(new AbsoluteLayout());

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("iAccess Authorization Agent GUI");
        setForeground(new java.awt.Color(51, 51, 51));
        setName("ClientFrame");
        setResizable(false);
        btn_StartNegotiations.setFont(new java.awt.Font("Dialog", 1, 15));
        btn_StartNegotiations.setForeground(new java.awt.Color(0, 51, 204));
        btn_StartNegotiations.setLabel("Submit Authorization Request");
        btn_StartNegotiations.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                btn_StartNegotiationsActionPerformed(evt);
            }
        });
        getContentPane().add(btn_StartNegotiations, new AbsoluteConstraints(430, 130, 350, 35));
        getContentPane().add(logo_background, new AbsoluteConstraints(635, 2, 150, 50));
        label1.setFont(new java.awt.Font("Dialog", 1, 14));
        label1.setForeground(new java.awt.Color(0, 51, 204));
        label1.setText("iAccess agent hostname/IP [port]:");
        getContentPane().add(label1, new AbsoluteConstraints(70, 22, 290, 25));


        Edit_ServerPort.setFont(new java.awt.Font("Dialog", 0, 16));
        Edit_ServerPort.setBackground(new java.awt.Color(255, 255, 255));
        Edit_ServerPort.setForeground(new java.awt.Color(0, 51, 204));
        Edit_ServerPort.setName("Edit_ServerPort");
        getContentPane().add(Edit_ServerPort, new AbsoluteConstraints(525, 20, 60, -1));

        label6.setFont(new java.awt.Font("Dialog", 1, 16));
        label6.setForeground(new java.awt.Color(0, 51, 204));
        label6.setText(":");
        getContentPane().add(label6, new AbsoluteConstraints(510, 21, -1, -1));

        
        label2.setFont(new java.awt.Font("Dialog", 1, 14));
        label2.setForeground(new java.awt.Color(0, 51, 204));
        label2.setText("Session time limit (sec):");
        getContentPane().add(label2, new AbsoluteConstraints(70, 60, 190, 30));

        label3.setFont(new java.awt.Font("Dialog", 1, 14));
        label3.setForeground(new java.awt.Color(0, 51, 204));
        label3.setText("Session time extension (sec):");
        getContentPane().add(label3, new AbsoluteConstraints(390, 60, 210, 30));

        Edit_STimeLimit.setBackground(new java.awt.Color(255, 255, 255));
        Edit_STimeLimit.setForeground(new java.awt.Color(0, 51, 204));
        Edit_STimeLimit.setFont(new java.awt.Font("Dialog", 0, 16));
        Edit_STimeLimit.setName("Edit_SessionTimeLimit");
        getContentPane().add(Edit_STimeLimit, new AbsoluteConstraints(270, 60, 60, -1));

        Edit_DSTimeExtension.setBackground(new java.awt.Color(255, 255, 255));
        Edit_DSTimeExtension.setForeground(new java.awt.Color(0, 51, 204));
        Edit_DSTimeExtension.setFont(new java.awt.Font("Dialog", 0, 16));
        Edit_DSTimeExtension.setName("Edit_DSTimeExtention");
        getContentPane().add(Edit_DSTimeExtension, new AbsoluteConstraints(620, 60, 60, -1)); 
        
        label7.setFont(new java.awt.Font("Dialog", 1, 14));
        label7.setForeground(new java.awt.Color(0, 51, 204));
        label7.setText("iAccess Monitor :: Outgoing Service Requests");
        getContentPane().add(label7, new AbsoluteConstraints(25, 148, 370, 22));
        
        TextArea_ClientWorkSpace.setBackground(new java.awt.Color(255, 255, 255));
        TextArea_ClientWorkSpace.setEditable(false);
        TextArea_ClientWorkSpace.setFont(new java.awt.Font("Dialog", 0, 12));
        TextArea_ClientWorkSpace.setForeground(new java.awt.Color(0, 51, 204));
        getContentPane().add(TextArea_ClientWorkSpace, new AbsoluteConstraints(20, 170, 760, 250) );
        
        labelServer.setFont(new java.awt.Font("Dialog", 1, 14));
        labelServer.setForeground(new java.awt.Color(0, 51, 204));
        labelServer.setText("iAccess Monitor :: Incoming Service Requests");
        getContentPane().add(labelServer, new AbsoluteConstraints(25, 426 , 370, 22));

        TextArea_ServerWorkSpace.setBackground(new java.awt.Color(255, 255, 255));
        TextArea_ServerWorkSpace.setEditable(false);
        TextArea_ServerWorkSpace.setFont(new java.awt.Font("Dialog", 0, 12));
        TextArea_ServerWorkSpace.setForeground(new java.awt.Color(0, 51, 204));
        getContentPane().add(TextArea_ServerWorkSpace, new AbsoluteConstraints(20, 450, 760, 220 ));        
        
        _label5.setFont(new java.awt.Font("Dialog", 1, 14));
        _label5.setForeground(new java.awt.Color(0, 51, 204));
        _label5.setText("Service Name:");
        getContentPane().add(_label5, new AbsoluteConstraints(70, 100, -1, -1));

        Edit_ServerAddress.setBackground(new java.awt.Color(255, 255, 255));
        Edit_ServerAddress.setForeground(new java.awt.Color(0, 51, 204));
        Edit_ServerAddress.setFont(new java.awt.Font("Dialog", 0, 16));
        Edit_ServerAddress.setName("Edit_ServerAddress");       
        Edit_ServerAddress.setText("Enter Server IP Here");
        getContentPane().add(Edit_ServerAddress, new AbsoluteConstraints(370, 20, 133, -1));
        
        _label8.setFont(new java.awt.Font("Dialog", 1, 14));
        _label8.setForeground(new java.awt.Color(0, 51, 204));
        _label8.setText("Service Action:");
        getContentPane().add(_label8, new AbsoluteConstraints(390, 100, 120, 30));

        Edit_ServiceName.setBackground(new java.awt.Color(255, 255, 255));
        Edit_ServiceName.setForeground(new java.awt.Color(0, 51, 204));
        Edit_ServiceName.setFont(new java.awt.Font("Dialog", 0, 16));
        Edit_ServiceName.setName("Edit_ServiceName");
        Edit_ServiceName.setText("");
        getContentPane().add(Edit_ServiceName, new AbsoluteConstraints(200, 100, 170, -1));

        Edit_ServiceAction.setBackground(new java.awt.Color(255, 255, 255));
        Edit_ServiceAction.setForeground(new java.awt.Color(0, 51, 204));
        Edit_ServiceAction.setFont(new java.awt.Font("Dialog", 0, 16));
        Edit_ServiceAction.setName("Edit_ServiceAction");
        Edit_ServiceAction.setText("");
        getContentPane().add(Edit_ServiceAction, new AbsoluteConstraints(520, 100, 160, -1));

        setBounds((screenSize.width-800)/2, 0, 800, 710);
    }

    private void btn_StartNegotiationsActionPerformed(java.awt.event.ActionEvent evt) 
    {
    	try {
                DisplayableSessionID = "0";
    		//session info update    		
    		long SessionTimeLimit = Long.parseLong(Edit_STimeLimit.getText().trim());    		
    		long TimeExtemsion = Long.parseLong(Edit_DSTimeExtension.getText().trim());
    		//Get Server IP Address
    		InetAddress IP_address = InetAddress.getByName(Edit_ServerAddress.getText().trim());
    		int ServerPort = Integer.parseInt(Edit_ServerPort.getText().trim());
                //Services Requests
    		String ServiceName = Edit_ServiceName.getText().trim();
            String ServiceAction = Edit_ServiceAction.getText().trim();

                // The real start of Authorization request (triggering a TN process)
                // If null then deny has been received, otherwise a digitally signed 
                // SAML Assertion is given. To be used for validation and enforcement 
                // by an application enforcement point (PEP). 
                
                //The lines below are only for testing purposes of server mode negotiations.
                //org.iaccess.TNProtocol.TRequest request = new org.iaccess.TNProtocol.TRequest(ServiceName, true, "not_used_in_this_mode", "not_used_in_this_mode");
                //java.util.Vector ad = TestDispatcher.getAccessDecisionInOKKAMMode(request);
                //System.out.println("IACCESS: Server mode credentials: " + ad.toString());
                
                String accessDecisionToken = TestDispatcher.startNewSessionClient(ServiceName,ServiceAction,SessionTimeLimit,TimeExtemsion,IP_address,ServerPort);
                if(accessDecisionToken!=null)
                {
                    //Check if the signed saml token contains exactly what your service request was.
                    //Then CONNECT TO A PEP FOR ENFORCING THE GRANT DECISION 
                }
                else
                {
                    //DO WHAT YOU DESIGN FOR DENY
                }
                //The following line is to stop the IACCESS Authorization service
                //TestDispatcher.stopServer();
		} catch (Exception e) 
                {
                    System.err.println("IACCESS: EXCEPTION: GUI: " + e);
		}
    }
    
    /**
     * Main function to run the iAccessDispatcher as an authorization service.
    */
    public static void main(String args[]) {
        // main/initial setup for running IACCESS (iAccessDispatcher) as an authorization service. 
        java.awt.EventQueue.invokeLater(
        new Runnable() {
            public void run() {
                UserInterface MyInterface = new UserInterface(); 
                //initialize the dispatcher (the real TN server and its underlying JBOSS socket functionality)
                MyInterface.TestDispatcher = new iAccessDispatcher(System.getProperty("user.dir") + File.separator + "config.xml", System.getProperty("user.dir") + File.separator + "log"+File.separator+"Default_iAccess_Log.log", MyInterface.OutputDataStream);                
                //Set defualt values in GUI components
                MyInterface.Edit_STimeLimit.setText(String.valueOf(TSessionInfo.SessionTimeLimit/1000));
                MyInterface.Edit_DSTimeExtension.setText(String.valueOf(TSessionInfo.DynamicSessionTimeExtension/1000));                            
                MyInterface.Edit_ServerAddress.setText(TSessionInfo.IPaddress);
                MyInterface.Edit_ServerPort.setText(String.valueOf(TSessionInfo.Default_TCPportOut));
                MyInterface.Edit_ServiceName.setText("");
                MyInterface.Edit_ServiceAction.setText("");
                //MyInterface.Edit_ServiceName.setText("record_amilburk");
                //MyInterface.Edit_ServiceAction.setText("read");
                MyInterface.setVisible(true);
            }
        });
    }
       

    protected void finalize() throws Throwable {
        System.exit(1);
        //TestDispatcher.stopServer();
	    //TestDispatcher.join();
    }
}
